/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#include "FakeEventTime.h"
#include "DetDesc/RunChangeIncident.h"

//-----------------------------------------------------------------------------
// Implementation file for class : FakeEventTime
//
// 2006-09-21 : Marco Clemencic
//-----------------------------------------------------------------------------

// Declaration of the Tool Factory
DECLARE_COMPONENT( FakeEventTime )


//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
FakeEventTime::FakeEventTime( const std::string& type,
                              const std::string& name,
                              const IInterface* parent )
  : base_class ( type, name , parent )
{
  declareInterface<IEventTimeDecoder>(this);

  // properties used to generate fake times
  declareProperty("StartTime",       m_startTime = 0);
  declareProperty("TimeStep",        m_timeStep  = 0);
  declareProperty("StartRun",        m_startRun  = 0);
  declareProperty("EventsPerRun",    m_eventsPerRun  = 0);
}

//=========================================================================
//  Initialization
//=========================================================================
StatusCode FakeEventTime::initialize ( ) {
  // base class initialization
  StatusCode sc = base_class::initialize();
  if (!sc.isSuccess()) return sc;

  MsgStream log( msgSvc(), name() );
  log << MSG::INFO << "Event times generated from " << m_startTime << " with steps of " << m_timeStep << endmsg;
  log << MSG::INFO << "Run numbers generated from " << m_startRun << " every " << m_eventsPerRun << " events" << endmsg;

  return sc;
}

StatusCode FakeEventTime::start() {
  StatusCode sc = base_class::start();

  m_incSvc = serviceLocator()->service("IncidentSvc", true);
  if ( !m_incSvc )
    throw GaudiException("Service [IncidentSvc] not found", name(),
                         StatusCode::FAILURE);

  return sc;
}

StatusCode FakeEventTime::stop() {
  m_incSvc.reset();
  return base_class::stop();
}

void FakeEventTime::i_increment() {
  m_startTime += m_timeStep;
  if (++m_evtCount == m_eventsPerRun) {
    m_evtCount = 0;
    m_startRun++;
  }
}

//=========================================================================
//  Return the time of current event
//=========================================================================
Gaudi::Time FakeEventTime::getTime ( ) const {

  // As a real EventTimeDecoder, the event time can be retrieved only in RUNNING
  // state, i.e. when the event can be loaded.
  if (FSMState() != Gaudi::StateMachine::RUNNING)
    return Gaudi::Time::epoch();

  // Here we should get the time from the EventDataSvc
  Gaudi::Time currentTime(m_startTime);
  if (m_evtCount == 0) {
    m_incSvc->fireIncident(RunChangeIncident(name(), m_startRun, currentTime));
  }

  // increment for the next event
  FakeEventTime *myPtr = const_cast<FakeEventTime *>(this);
  myPtr->i_increment();

  return currentTime;
}
//=============================================================================
