/*****************************************************************************\
* (c) Copyright 2000-2019 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef LHCBMATH_POWER_HPP 
#define LHCBMATH_POWER_HPP 1
// ============================================================================
// Include files
// ============================================================================

// ============================================================================
/** @file Power.hpp
 *
 *
 *  @author Juan PALACIOS juan.palacios@cern.ch
 *  @date 2006-10-26 
 */
// ============================================================================

namespace Gaudi 
{
  namespace Math 
  {
    
    /** @fn Gaudi::Math::pow
     *  Template metafunction to calculate integer powers of integer
     *  and floating point numbers.
     *
     *  @code 
     *   // calculate 50.**10
     *   const double result = Gaudi::Math::pow<double, 10> ( 50. ) ;
     *
     *  @code 
     *
     *  @author Juan PALACIOS juan.palacios@cern.ch
     *  @date 2006-10-26 
     */


    template<typename TYPE, int N>
    struct Power {
      static inline TYPE pow ( TYPE __x ) { 
        return Power<TYPE, N-1>::pow(__x)*__x;
      }
    };
    
    template<typename TYPE>
    struct Power<TYPE, 1> {      
      static inline TYPE pow ( TYPE __x ) { return __x; }
    };

    template<typename TYPE>
    struct Power<TYPE, 0> {
      static inline TYPE pow ( TYPE __x ) { return 1; }
    };

    template<typename TYPE, int N, bool C>
    struct ReturnPolicy {
      typedef TYPE RETURN_TYPE;
    };
    
    template<typename TYPE, int N>
    struct ReturnPolicy<TYPE,N,false> {
      typedef double RETURN_TYPE;
    };

    template<typename TYPE, int N>
    struct InvPower
    {
      static inline typename ReturnPolicy<TYPE, N, (N>=0) >::RETURN_TYPE pow( TYPE __x )
      {
        return 1./Power<TYPE, -1*N>::pow(__x);
      }  
    };

    template <typename TYPE, int N, bool C>
    struct ImplementationSwitch {
      typedef Gaudi::Math::Power<TYPE, N>  powImpl;
    };
    
    template <typename TYPE, int N>
    struct ImplementationSwitch<TYPE, N, false> {
      typedef Gaudi::Math::InvPower<TYPE, N>  powImpl;
    };
    
    template<typename TYPE, int N>
    static inline typename ReturnPolicy<TYPE, N, (N>=0) >::RETURN_TYPE pow ( TYPE __x ) 
    {
      return ImplementationSwitch< TYPE, N, (N>=0) >::powImpl::pow(__x);
    };
    

  } // end of namespace Math
  
} // end of namespace Gaudi

// ============================================================================
// The END 
// ============================================================================
#endif // LHCBMATH_POWER_HPP
// ============================================================================
