/*****************************************************************************\
 * (c) Copyright 2018 CERN for the benefit of the LHCb Collaboration           *
 *                                                                             *
 * This software is distributed under the terms of the GNU General Public      *
 * Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
 *                                                                             *
 * In applying this licence, CERN does not waive the privileges and immunities *
 * granted to it by virtue of its status as an Intergovernmental Organization  *
 * or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

// Include file
// local
#include "GenFSRJson.h"

// from Event
#include "Event/CrossSectionsFSR.h"
#include "Event/GenCountersFSR.h"
#include "Event/GenFSR.h"

// From GaudiKernel
#include "GaudiKernel/Time.h"

// other libraries
#include <cmath>
#include <fstream>
#include <iostream>
#include <string>

#include "boost/property_tree/json_parser.hpp"
#include "boost/property_tree/ptree.hpp"

using boost::property_tree::ptree;

#define script_version "20181108"

//-----------------------------------------------------------------------------
// Implementation file for class : GenFSRJson
//
// 2018-06-26 : Davide Fazzini
//
//
// N.B: The genCounters are identified by means of a key number.
//      The complete list of all the counters is here: /Event/GenEvent/xml/GenCountersFSR.xml
//      Since counters belonging to the same class have consecutive keys, the methods used by
//      this service cut on the key variable in order to select the correct counter class.
//-----------------------------------------------------------------------------

namespace {
  std::string getCurrentTime() { return Gaudi::Time::current().format( true ); }

  const std::string getEvtTypeDesc( int evtType ) {
    std::string description;
    std::string path_decFile = System::getEnv( "DECFILESROOT" );

    if ( path_decFile != "" ) {
      std::string  evtType_str = dynamic_cast<std::ostringstream*>(&(std::ostringstream() << evtType))->str();
      std::string  path_desc   = path_decFile + "/options/" + evtType_str + ".py";
      std::fstream file_desc( path_desc.c_str(), std::fstream::in );
      std::string  line_ref  = "$DECFILESROOT/dkfiles/";
      std::string  line_stop = ".dec\"";
      std::string  line;

      while ( getline( file_desc, line ) ) {
        std::size_t pos_ref = line.find( line_ref );

        if ( pos_ref != std::string::npos ) {
          std::size_t pos_end = line.find( line_stop );
          pos_ref             = pos_ref + line_ref.length();
          int count           = pos_end - pos_ref;
          description         = line.substr( pos_ref, count );
          break;
        }
      }
      file_desc.close();
    }
    return description;
  }

  //=============================================================================
  //  Add generator level counters in the json file
  //=============================================================================

  ptree writeGeneratorCounters( const LHCb::GenFSR& genFSR ) {
    LHCb::CrossSectionsFSR             crossFSR;
    std::map<std::string, std::string> mapCross = crossFSR.getFullNames();

    LHCb::GenFSR::GeneratorCounter::iterator icounter;    
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    
    ptree array_tree;

    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;

      if ( key > LHCb::GenCountersFSR::CounterKeyToType("bAndcAcc") ) continue;

      if ( key == LHCb::GenCountersFSR::CounterKeyToType("BeforeFullEvt") ||
           key == LHCb::GenCountersFSR::CounterKeyToType("AfterFullEvt") )
        continue;

      longlong before = genFSR.getDenominator( key );
      if ( before == 0 ) continue;

      double C = ( key >= LHCb::GenCountersFSR::CounterKeyToType("OnebGen")
                   ? genFSR.getCrossSectionInfo( LHCb::CrossSectionsFSR::CrossSectionKeyToType("MBCrossSection") ).second
                       : 1 );

      longlong after =
        ( key == LHCb::GenCountersFSR::CounterKeyToType("EvtGenerated") ? genFSR.getGenCounterInfo( key + 1 ).second
                                                                  : icounter->second.second );
      if ( after == 0 ) continue;

      bool flag = ( key != LHCb::CrossSectionsFSR::CrossSectionKeyToType("MeanNoZeroPUInt") &&
                    key != LHCb::CrossSectionsFSR::CrossSectionKeyToType("MeanPUInt") &&
                    key != LHCb::CrossSectionsFSR::CrossSectionKeyToType("MeanPUIntAcc") );

      double      fraction = genFSR.getEfficiency( after, before, C );
      double      error    = genFSR.getEfficiencyError( after, before, C, flag );
      std::string fullName = mapCross[LHCb::CrossSectionsFSR::CrossSectionKeyToString( key )];

      ptree counter_tree;
      counter_tree.put( "descr", fullName );
      counter_tree.put( "type", "counter" );
      counter_tree.put( "value", fraction );
      counter_tree.put( "error", error );

      array_tree.push_back( std::make_pair( "", counter_tree ) );
    }

    return array_tree;
  }

  //=============================================================================
  //  Add generator level cut efficiencies in the xml file
  //=============================================================================

  ptree writeCutEfficiencies( const LHCb::GenFSR& genFSR ) {
    LHCb::CrossSectionsFSR             crossFSR;
    std::map<std::string, std::string> mapCross = crossFSR.getFullNames();

    LHCb::GenFSR::GeneratorCounter::iterator icounter;    
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    
    ptree array_tree;

    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;

      if ( key != LHCb::GenCountersFSR::CounterKeyToType("AfterFullEvt") &&
           key != LHCb::GenCountersFSR::CounterKeyToType("AfterLevelCut") &&
           key != LHCb::GenCountersFSR::CounterKeyToType("EvtSignal") &&
           key != LHCb::GenCountersFSR::CounterKeyToType("EvtantiSignal") &&
           key != LHCb::GenCountersFSR::CounterKeyToType("AfterPCut") &&
           key != LHCb::GenCountersFSR::CounterKeyToType("AfterantiPCut") )
        continue;

      longlong before = genFSR.getDenominator( key );
      if ( before == 0 ) continue;
      longlong after = icounter->second.second;
      if ( after == 0 ) continue;

      double fraction = genFSR.getEfficiency( after, before );
      double error    = genFSR.getEfficiencyError( after, before );

      std::string fullName = mapCross[LHCb::CrossSectionsFSR::CrossSectionKeyToString( key )];

      ptree efficiency_tree;

      efficiency_tree.put( "descr", fullName );
      efficiency_tree.put( "type", "counter" );
      efficiency_tree.put( "value", fraction );
      efficiency_tree.put( "error", error );

      array_tree.push_back( std::make_pair( "", efficiency_tree ) );
    }

    return array_tree;
  }

  //=============================================================================
  //  Add generated and accepted hadron counters in the json file
  //=============================================================================

  ptree writeHadronCounters( const LHCb::GenFSR& genFSR ) {
    ptree array_tree;

    LHCb::CrossSectionsFSR             crossFSR;
    std::map<std::string, std::string> mapCross = crossFSR.getFullNames();

    LHCb::GenFSR::GeneratorCounter::iterator icounter;    
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;

      if ( ( key >= LHCb::GenCountersFSR::CounterKeyToType("B0Gen") && key <= LHCb::GenCountersFSR::CounterKeyToType("D2starAcc") ) ) {
        longlong before = genFSR.getDenominator( key );
        if ( before == 0 ) continue;
        longlong after = icounter->second.second;
        if ( after == 0 ) continue;

        double fraction = genFSR.getEfficiency( after, before );
        double error    = genFSR.getEfficiencyError( after, before );

        std::string fullName = mapCross[LHCb::CrossSectionsFSR::CrossSectionKeyToString( key )];

        ptree counter_tree;
        counter_tree.put( "descr", fullName );
        counter_tree.put( "type", "counter" );
        counter_tree.put( "value", fraction );
        counter_tree.put( "error", error );

        array_tree.push_back( std::make_pair( "", counter_tree ) );
      }
    }

    return array_tree;
  }

  //=============================================================================
  //  Add generator cross-sections in the json file
  //=============================================================================

  ptree writeGeneratorCrossSections( const LHCb::GenFSR& genFSR ) {
    ptree array_tree;

    LHCb::GenFSR::CrossSection::iterator icross;    
    LHCb::GenFSR::CrossSection crossSections = genFSR.getCrossSections();

    for(icross = crossSections.begin(); icross != crossSections.end(); icross++) {
      std::string name = icross->second.first;
      name.erase( name.size() - 2 );
      double value = icross->second.second;

      if ( value == 0 || std::isnan( value ) != 0 ) continue;

      ptree cross_tree;
      cross_tree.put( "descr", name );
      cross_tree.put( "type", "cross-section" );
      cross_tree.put( "value", value );

      array_tree.push_back( std::make_pair( "", cross_tree ) );
    }

    return array_tree;
  }

  //=============================================================================
  //  Add global statistics in the json file
  //=============================================================================

  ptree writeGlobalStat( const LHCb::GenFSR& genFSR ) {
    longlong countEvtGen = 0, countEvtAcc = 0, countIntGen = 0, countIntAcc = 0;

    ptree array_tree;

    LHCb::GenFSR::GeneratorCounter::iterator icounter;    
    LHCb::GenFSR::GeneratorCounter genCounters = genFSR.getGenCounters();
    
    for(icounter = genCounters.begin(); icounter != genCounters.end(); icounter++) {
      int key = icounter->first;
      if ( key > LHCb::GenCountersFSR::CounterKeyToType("IntAccepted") ) continue;

      if ( key == LHCb::GenCountersFSR::CounterKeyToType("EvtGenerated") ) 
        countEvtGen = icounter->second.second;
      else if ( key == LHCb::GenCountersFSR::CounterKeyToType("IntGenerated") ) 
        countIntGen = icounter->second.second;
      else if ( key == LHCb::GenCountersFSR::CounterKeyToType("EvtAccepted") ) 
        countEvtAcc = icounter->second.second;      
      else if ( key == LHCb::GenCountersFSR::CounterKeyToType("IntAccepted") ) 
        countIntAcc = icounter->second.second;
    }

    ptree event_tree, interaction_tree;

    event_tree.put( "descr", "Number of accepted events/generated events" );
    event_tree.put( "type", "fraction" );
    event_tree.put( "numer", countEvtAcc );
    event_tree.put( "denom", countEvtGen );

    interaction_tree.put( "descr", "Number of interactions in accepted events/generated interactions" );
    interaction_tree.put( "type", "fraction" );
    interaction_tree.put( "numer", countIntAcc );
    interaction_tree.put( "denom", countIntGen );

    array_tree.push_back( std::make_pair( "", event_tree ) );
    array_tree.push_back( std::make_pair( "", interaction_tree ) );

    return array_tree;
  }
} // namespace

// Instantiation of a static factory class used by clients to create
// instances of this service
DECLARE_ALGORITHM_FACTORY( GenFSRJson )

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
GenFSRJson::GenFSRJson( const std::string& name, 
                        ISvcLocator* pSvcLocator) 
: GaudiAlgorithm(name, pSvcLocator),
  m_fileRecordSvc(NULL),
  m_navigatorTool(NULL)
{
  declareProperty("FileRecordLocation", m_fileRecordName = "/FileRecords");
  declareProperty("FSRName"           , m_FSRName = "/GenFSR");
  declareProperty("jsonOutputLocation", m_jsonOutputLocation = "");
  declareProperty("jsonOutputName"    , m_jsonOutputName = "GenerationFSRnew.json");
  declareProperty("prodID"            , m_prodID = "");
  declareProperty("appConfigVersion"  , m_appConfigVersion = "");
  declareProperty("appConfigFile"     , m_appConfigFile = "");
  declareProperty("gaussVersion"      , m_gaussVersion = "");
  declareProperty("simCond"           , m_simCond = "");
  declareProperty("dddb"              , m_dddb = "");
}

//=============================================================================
//  Initialization
//=============================================================================

StatusCode GenFSRJson::initialize() {
  StatusCode sc = GaudiAlgorithm::initialize(); // must be executed first

  if ( sc.isFailure() ) return sc; // error prinsted already by Service
  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Initialize" << endmsg;

  // get the File Records service
  m_fileRecordSvc = svc<IDataProviderSvc>("FileRecordDataSvc", true);  

  m_navigatorTool = tool<IFSRNavigator>( "FSRNavigator", "FSRNavigator" );

  return sc;
}

//=============================================================================
// Main execution
//=============================================================================
StatusCode GenFSRJson::execute() {
  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Execute" << endmsg;

  return StatusCode::SUCCESS;
}

//=============================================================================
//  Finalize
//=============================================================================
StatusCode GenFSRJson::finalize() {
  if ( msgLevel( MSG::DEBUG ) ) debug() << "==> Finalize" << endmsg;

  GenFSRJson::printFSR();

  return GaudiAlgorithm::finalize(); // must be called after all other actions
}

//=============================================================================
//  Printthe GenFSR in a file json
//=============================================================================

void GenFSRJson::printFSR() {

  // make an inventory of the FileRecord store
  std::vector<std::string> addresses = m_navigatorTool->navigate( m_fileRecordName, m_FSRName );

  for(std::vector< std::string >::iterator iAddr=addresses.begin(); iAddr!=addresses.end(); iAddr++) {
    DataObject* obj = NULL;
    std::string genRecordAddress = *iAddr;
    StatusCode  sc  = m_fileRecordSvc->retrieveObject( genRecordAddress, obj );

    if ( !sc.isSuccess() ) {
      error() << "Unable to retrieve object '" << genRecordAddress << "'" << endmsg;
      continue;
    }

    LHCb::GenFSR* genFSR = dynamic_cast<LHCb::GenFSR*>( obj );
    if ( genFSR == NULL ) {
      warning() << "genFSR record not found!" << endmsg;
      if ( msgLevel( MSG::DEBUG ) ) debug() << genRecordAddress << " not found" << endmsg;
    } else {
      int evtType = genFSR->getSimulationInfo( "evtType", 0 );
      int njobs   = genFSR->getSimulationInfo( "nJobs", 0 );

      std::string genMethod   = genFSR->getSimulationInfo( "generationMethod", "" );
      std::string genName     = genFSR->getSimulationInfo( "hardGenerator", "" );
      std::string decFiles    = genFSR->getSimulationInfo( "decFiles", "" );
      std::string evtType_str = dynamic_cast<std::ostringstream*>(&(std::ostringstream() << evtType))->str();
      std::string njobs_str   = dynamic_cast<std::ostringstream*>(&(std::ostringstream() << njobs))->str();
      std::string time        = getCurrentTime();

      // open the new file
      if ( msgLevel( MSG::DEBUG ) ) debug() << "write to file: " + m_jsonOutputName << endmsg;
      std::ofstream jsonOutput( (m_jsonOutputLocation + m_jsonOutputName).c_str(), std::fstream::out );

      if ( jsonOutput.is_open() ) {
        if ( msgLevel( MSG::DEBUG ) )
          debug() << "Json output: " + m_jsonOutputLocation + m_jsonOutputName + " created." << endmsg;

        ptree       main_tree;
        std::string evtDesc = getEvtTypeDesc( evtType );

        main_tree.put( "APPCONFIG_file", m_appConfigFile );
        main_tree.put( "APPCONFIG_version", m_appConfigVersion );
        main_tree.put( "evtType", evtType_str );
        main_tree.put( "evtTypeDesc", evtDesc );
        main_tree.put( "DecFiles_version", decFiles );
        main_tree.put( "Hard Generator", genName );
        main_tree.put( "Gauss_version", m_gaussVersion );
        main_tree.put( "DDDB", m_dddb );
        main_tree.put( "SIMCOND", m_simCond );
        main_tree.put( "genTimeStamp", time );

        ptree gencounters_array    = writeGeneratorCounters( *genFSR );
        ptree hadroncounters_array = writeHadronCounters( *genFSR );
        ptree cross_array          = writeGeneratorCrossSections( *genFSR );
        ptree efficiencies_array   = writeCutEfficiencies( *genFSR );
        ptree stats_array          = writeGlobalStat( *genFSR );

        main_tree.add_child( "InteractionCounters", gencounters_array );
        main_tree.add_child( "HadronCounters", hadroncounters_array );
        main_tree.add_child( "HardGeneratorCrossSection", cross_array );
        main_tree.add_child( "SignalCounters", efficiencies_array );
        main_tree.add_child( "globStat", stats_array );

        main_tree.put( "prodID", m_prodID );
        main_tree.put( "nb_jobs", njobs_str );
        main_tree.put( "script_version", script_version );

        write_json( jsonOutput, main_tree, false );
        jsonOutput.close();
      } else if ( msgLevel( MSG::DEBUG ) )
        debug() << "The output file was not opened correctly" << endmsg;
    }
  }
}
